from re import template
import questionary
import yaml
import os
import boto3
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.validation import Validator, ValidationError
from os.path import exists
from jinja2 import Template, meta, Environment
from jinja2schema import infer_from_ast, model
from questionary import Style
from botocore.config import Config
from botocore.exceptions import ClientError

class FileExistsValidator(Validator):
    def validate(self, document):
        if (exists(document.text) == False):
            raise ValidationError(message="Please enter a valid path to file",
                                  cursor_position=len(document.text))

def get_file(message):
    return questionary.text(message, completer=PathCompleter() ,validate=FileExistsValidator, validate_while_typing=False).ask()


def get_env_spec_inputs(spec_yaml):
    inputs = {}
    inputs["environment"] = spec_yaml["spec"]

    inputs["environment"]["name"] = questionary.text("Enter environment name", 
                                                      validate=lambda text: len(text) > 0,
                                                      validate_while_typing=False).ask()

    return inputs


def get_svc_spec_inputs(spec_yaml, template):
    options = []
    if "pipeline" in spec_yaml:
        options.append("pipeline")
    
    for instance in spec_yaml["instances"]:
        options.append(instance["name"])

    all_variables = get_all_variables(template)

    selected_instance = questionary.select("Which instance would you like to compile?", choices=options).ask()

    if selected_instance == "pipeline":
        inputs = {}
        inputs["pipeline"] = {}
        inputs['pipeline']['inputs'] = spec_yaml['pipeline']
        instance_inputs = []
        for instance in spec_yaml["instances"]:
            input = {}
            input["name"] = instance["name"]
            input["inputs"] = instance["spec"]
            instance_inputs.append(input)
        inputs["service_instances"] = instance_inputs
        return inputs
    else:
        for instance in spec_yaml["instances"]:
            if instance["name"] == selected_instance:
                inputs = {}
                inputs["environment"] = {}
                inputs["environment"]["name"] = instance["environment"]
                if "environment" in all_variables:
                    environment_source = questionary.select("It looks like your template relies on some variables from a Proton Environment. How would you like to provide those values",
                                                            choices=["Retrieve outputs from AWS Proton for environment: " + instance["environment"],
                                                                     "Provide inputs from inferred schema"]).ask()
                    if environment_source == "Provide inputs from inferred schema":
                        inputs["environment"] = get_raw_input(all_variables["environment"], "environment")
                    else:
                        proton_client = boto3.client('proton', config=Config(region_name='us-east-1'))
                        outputs = proton_client.list_environment_outputs(environmentName=instance["environment"])["outputs"]
                        output_dict = {}
                        for output in outputs:
                            output_dict[output["key"]] = output["valueString"]
                        inputs["environment"]["outputs"] = output_dict

                inputs["service_instance"] = {}
                inputs["service_instance"]["name"] = selected_instance
                inputs["service_instance"]["inputs"] = instance["spec"]
                return inputs

def get_inputs_from_spec(template):
    spec_file = get_file("Enter path to spec file")
    spec_yaml = None
    inputs = None
    with open(spec_file, "r") as spec:
        spec_yaml = yaml.safe_load(spec)
    if spec_yaml["proton"] == "EnvironmentSpec":
        return get_env_spec_inputs(spec_yaml)
    elif spec_yaml["proton"] == "ServiceSpec":
        return get_svc_spec_inputs(spec_yaml, template)
    else:
        print("Unexpected spec type, unable to get inputs: " + spec_yaml["proton"])
        raise


def get_raw_input(all_variables, prefix=""):
    inputs = {}
    for key, value in all_variables.items():
        cur_var = key
        if prefix != "":
            cur_var = prefix + "." + cur_var
        if type(value) is model.Dictionary:
            inputs[key] = get_raw_input(value, cur_var)
        elif type(value) is model.Scalar:
            inputs[key] = questionary.text("Enter value for " + cur_var).ask()
        else:
            print("Unexpected variable type, unable to infer schema: " + type(value))
            raise 
    return inputs
    


def get_inputs(template):
    input_style = questionary.select(
        "How would you like to provide your inputs?",
        choices=["Proton Spec File", "Scan jinja template for variables"]
    ).ask()

    if (input_style == "Proton Spec File"):
        return get_inputs_from_spec(template)
    elif (input_style == "Scan jinja template for variables"):
        all_variables = get_all_variables(template)
        return get_raw_input(all_variables)


def render_jinja(jinja_file, inputs):
    with open(jinja_file, 'r') as file_:
        jinja_template = Template(file_.read())
    return jinja_template.render(inputs)


def get_all_variables(jinja_file):
    with open(jinja_file, 'r') as file_:
        env = Environment()
        ast = env.parse(file_.read())
    return infer_from_ast(ast)


def compile():
    cfn_template = get_file("Enter path to cfn template")
    inputs = get_inputs(cfn_template)
    rendered_jinja = render_jinja(cfn_template, inputs)
    print(rendered_jinja)
    next_choice = questionary.select(
        "You've got some more options if you'd like, or you can exit",
        choices=["Exit", "Validate CFN", "Deploy CFN"]
    ).ask()

    if next_choice == "Validate CFN":
        try:
            cfn_client = boto3.client('cloudformation', config=Config(region_name='us-east-1'))
            cfn_client.validate_template(TemplateBody=rendered_jinja)
            print("Valid CFN!")
        except ClientError as e:
            print(e)


def check_if_in_template():
    if "schema" in os.listdir('.'):
        return True
    return False


def cli_entry():
    # if (check_if_in_template()):
        # questionary.confirm("It looks like you're running me inside of a Proton Template directory, do you want to work with this template?").ask()
    execution_mode = questionary.select(
        "What would you like to do?",
        choices=["Compile", "Bundle Template"]
    ).ask()

    if (execution_mode == "Compile"):
        compile()
    elif (execution_mode == "Bundle Template"):
        print("Not yet supported")


def main():
    cli_entry()


if __name__ == "__main__":
    main()